#include <iostream>

#include "doctest.h"
#include "lib/hex.cpp"

TEST_CASE("hex_arithmetic") {
    CHECK( Hex(4, -10, 6) == (Hex(1, -3, 2) + Hex(3, -7, 4)) );
    CHECK( Hex(-2, 4, -2) == (Hex(1, -3, 2) - Hex(3, -7, 4)) );
}


TEST_CASE("hex_direction") {
    CHECK( Hex(0, -1, 1) ==  hex_direction(2) );
}


TEST_CASE("hex_neighbor") {
    CHECK( hex_neighbor(Hex(1, -2, 1), 2) ==  Hex(1, -3, 2));
}


TEST_CASE("hex_diagonal") {
	CHECK( hex_diagonal_neighbor(Hex(1, -2, 1), 3) == Hex(-1, -1, 2));
}

TEST_CASE("hex_distance") {
	CHECK(hex_distance(Hex(3, -7, 4), Hex(0, 0, 0)) == 7);
}

TEST_CASE("rotate_right") {
	CHECK( Hex(1, -3, 2).rotate_right() ==  Hex(3, -2, -1) );
}

TEST_CASE("rotate_left") {
	CHECK( Hex(1, -3, 2).rotate_left() ==  Hex(-2, -1, 3) );
}

TEST_CASE("hex_round") {
    FractionalHex a = FractionalHex(0.0, 0.0, 0.0);
    FractionalHex b = FractionalHex(1.0, -1.0, 0.0);
    FractionalHex c = FractionalHex(0.0, -1.0, 1.0);

    CHECK( Hex(5, -10, 5) == hex_round(hex_lerp(FractionalHex(0.0, 0.0, 0.0), FractionalHex(10.0, -20.0, 10.0), 0.5)) );
    CHECK( hex_round(a) ==  hex_round(hex_lerp(a, b, 0.499)) );
    CHECK( hex_round(b) == hex_round(hex_lerp(a, b, 0.501)));
    CHECK( hex_round(a) == hex_round(FractionalHex(a.q * 0.4 + b.q * 0.3 + c.q * 0.3, a.r * 0.4 + b.r * 0.3 + c.r * 0.3, a.s * 0.4 + b.s * 0.3 + c.s * 0.3)));
    CHECK( hex_round(c) == hex_round(FractionalHex(a.q * 0.3 + b.q * 0.3 + c.q * 0.4, a.r * 0.3 + b.r * 0.3 + c.r * 0.4, a.s * 0.3 + b.s * 0.3 + c.s * 0.4)));
}

//TEST_CASE("hex_linedraw") {
//	Hex[] path = {Hex(0, 0, 0), Hex(0, -1, 1), Hex(0, -2, 2), Hex(1, -3, 2), Hex(1, -4, 3), Hex(1, -5, 4)};
//	CHECK( path == hex_linedraw(Hex(0, 0, 0), Hex(1, -5, 4)) );
//}

TEST_CASE("layout_flat") {
    Hex h = Hex(3, 4, -7);
    Layout flat = Layout(layout_flat, Point(10.0, 15.0), Point(35.0, 71.0));
    CHECK( h == hex_round(pixel_to_hex(flat, hex_to_pixel(flat, h))) );
}

TEST_CASE("layout_pointy") {
    Hex h = Hex(3, 4, -7);
    Layout pointy = Layout(layout_pointy, Point(10.0, 15.0), Point(35.0, 71.0));
    CHECK(h == hex_round(pixel_to_hex(pointy, hex_to_pixel(pointy, h))));
}

TEST_CASE("offset_roundtrip") {
    Hex a = Hex(3, 4, -7);
    OffsetCoord b = OffsetCoord(1, -3);
    CHECK(a == qoffset_to_cube(EVEN, qoffset_from_cube(EVEN, a)));
    CHECK(b == qoffset_from_cube(EVEN, qoffset_to_cube(EVEN, b)));

    CHECK(a == qoffset_to_cube(ODD, qoffset_from_cube(ODD, a)));
    CHECK(b == qoffset_from_cube(ODD, qoffset_to_cube(ODD, b)));

    CHECK(a == roffset_to_cube(EVEN, roffset_from_cube(EVEN, a)));
    CHECK(b == roffset_from_cube(EVEN, roffset_to_cube(EVEN, b)));

    CHECK(a == roffset_to_cube(ODD, roffset_from_cube(ODD, a)));
    CHECK(b == roffset_from_cube(ODD, roffset_to_cube(ODD, b)));
}

TEST_CASE("offset_from_cube") {
	CHECK(OffsetCoord(1, 3) == qoffset_from_cube(EVEN, Hex(1, 2, -3)));
	CHECK(OffsetCoord(1, 2) == qoffset_from_cube(ODD, Hex(1, 2, -3)));
}

TEST_CASE("offset_to_cube") {
    CHECK(Hex(1, 2, -3) == qoffset_to_cube(EVEN, OffsetCoord(1, 3)));
    CHECK(Hex(1, 2, -3) == qoffset_to_cube(ODD, OffsetCoord(1, 2)));
}

